/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import "IDLDefinition_Private.h"
#import "IDLStruct.h"
#import <Foundation/Foundation.h>
#import "IDLStructMember.h"
#import "IDLEncapsulation.h"
#import "IDLLink.h"
#import "IDLDefinitionScope.h"
#import "IDLObject.h"
#import "IDLException.h"

@interface NSObject(IDLObject)
- initWithIDLType:(id<IDLDefinition>)type;
- (id<IDLDefinition>) idlType;
@end

@implementation IDLStruct(CORBACoding)

static Class _objectClass = nil;
static BOOL logEnabled = NO;

+ (void) setLogEnabled:(BOOL) flag
{
	logEnabled = flag?YES:NO;
}

+ (Class) defaultObjectClass
{
	if(!_objectClass)
		_objectClass = [IDLObject class];
	return _objectClass;
}

+ (void) setDefaultObjectClass:(Class)class
{
	_objectClass = class;
}

- (void)encodeObject:anObject withCoder:(id <CORBACoder>)anEncoder
{
    NSEnumerator* en = [_members objectEnumerator];
    IDLStructMember* member;
    while(member = [en nextObject]){
        id def = [member type];
		id obj = nil;
		@try{
			obj = [anObject valueForKey:[member name]]; // Using KeyValueCoding to access any object
		}@catch(NSException *localException){
			if([[localException name]isEqualToString:NSUndefinedKeyException]){
				ORBLogUndefinedKeyError(self, anObject, [member name]);
				if(![IDLDefinition allowsUndefinedKeys])
					[localException raise];
			}else
				[localException raise];
		}
        [def encodeObject:obj withCoder:anEncoder];
    }
}

- (void) decodeObjectState:(id*)value withCoder:(id <CORBADecoder>)coder
{
    NSEnumerator* en = [_members objectEnumerator];
    IDLStructMember* member;
	int pos0 = 0;
    while(member = [en nextObject]){
        id def = [member type];
        id obj = nil;
		if(logEnabled)
			pos0 = [coder cursorPosition]; 
        [def decodeObject:&obj withCoder:coder];
		@try{
			[*value setValue:obj forKey:[member name]];
			if(logEnabled)
				NSLog(@"valuebox %@: decoded '%@ = 0x%x' for '0x%x' range %d.%d\n", 
					[self name], [member name], obj, *value, pos0, [coder cursorPosition]-pos0);
		}@catch(NSException *localException){
			if([[localException name]isEqualToString:NSUndefinedKeyException]){
				ORBLogUndefinedKeyError(self, *value, [member name]);
				if(![IDLDefinition allowsUndefinedKeys])
					[localException raise];
			}else
				[localException raise];
		}
    }
}

- (void) decodeObject:(id*)objectPtr withCoder:(id <CORBADecoder>)anEncoder usingClass:objectClass
{
    NSEnumerator* en = [_members objectEnumerator];
    IDLStructMember* member;
    id dict;
	int pos0 = 0;
	if([objectClass instancesRespondToSelector:@selector(initWithIDLType:)])
		dict = [[[objectClass alloc]initWithIDLType:self]autorelease];
	else
		dict = [[[objectClass alloc]init]autorelease];
    while(member = [en nextObject]){
        id def = [member type];
        id obj = nil;
		if(logEnabled)
			pos0 = [anEncoder cursorPosition]; 
        [def decodeObject:&obj withCoder:anEncoder];
		@try{
			[dict setValue:obj forKey:[member name]];
			if(logEnabled)
				NSLog(@"struct %@: decoded '%@ = 0x%x' for '0x%x' range %d.%d\n", 
					[self name], [member name], obj, dict, pos0, [anEncoder cursorPosition]-pos0);
		}@catch(NSException *localException){
			if([[localException name]isEqualToString:NSUndefinedKeyException]){
				ORBLogUndefinedKeyError(self, dict, [member name]);
				if(![IDLDefinition allowsUndefinedKeys])
					[localException raise];
			}else
				[localException raise];
		}
    }
    *objectPtr = dict;
}
- (void)decodeObject:(id *)objectPtr withCoder:(id <CORBADecoder>)anEncoder
{
    id objectClass = _className? NSClassFromString(_className) : [[self class] defaultObjectClass];
    if(!objectClass)
        objectClass = [[self class] defaultObjectClass];
    [self decodeObject:objectPtr withCoder:anEncoder usingClass:objectClass];
}

- (const char*) objCType
{
    return @encode(id);
}
- (unsigned) argLength
{
    return sizeof(id);
}
- (unsigned) returnValueLength
{
    return sizeof(id);
}

- argumentObjectFromInvocation:(NSInvocation*)inv atIndex:(int)n
{
    // Struct is mapped to NSDictionary
    id dict = nil;
    [inv getArgument:&dict atIndex:n];
    return dict;
}

- argumentObjectOfType:(const char*)argType fromPointer:(void*)ptr
{
    id dict = *((id*)ptr);
    return dict;
}

- (void) setValue:value ofType:(const char*)argType forPointer:(void*)ptr
{
    *((id*)ptr) = value;
}
- (void) setValue:value forInvocation:(NSInvocation*)inv atIndex:(int)n
{
    id _val = value;
    [inv setArgument:&_val atIndex:n];
}
- (void) setReturnValue:value forInvocation:(NSInvocation*)inv
{
    id _val = value;
    [inv setReturnValue:&_val];
}
- returnValueFromInvocation:(NSInvocation*)inv
{
    id _val = nil;
    [inv getReturnValue:&_val];
    return _val;
}

- (void) _encodeWithCORBACoder:(id <CORBACoder>)coder
{
    NSEnumerator* en = [_members objectEnumerator];
    IDLStructMember* member;
    [coder encodeString:[self typeId]];
    [coder encodeString:[self name]];
    [coder encodeLong:[_members count]];
    while(member = [en nextObject]){
        IDLDefinition* def = (IDLDefinition*)[member type];
        [coder encodeString:[member name]];
        [def encodeWithCORBACoder:coder];
    }
}

- (void) encodeWithCORBACoder:(id <CORBACoder>)coder
{
    long location = [coder locationForDefinition:self];
    if(location != 0 && [IDLDefinition allowsIndirection]) // already encoded -> encode indirection
        [coder encodeIndirection:location];
    else{
        IDLEncapsulation* stream = [IDLEncapsulation encapsulation];
        [coder encodeLong:[self kind]];
        [stream setParentStream:coder location:[coder location]];
        [stream registerLocation:-2*sizeof(long) forDefinition:self];
        [self _encodeWithCORBACoder:stream];
        [coder encodeData:[stream data]];
    }
}

+ _definitionFromCoder:(IDLEncapsulation*)coder
{
    IDLLink* link = [IDLLink linkWithDefinition:nil];
    NSMutableArray* array = [NSMutableArray array];
    NSString *str;
    NSString* typeId;
    IDLStruct* obj = nil;
    IDLStruct* existingObj = nil;
    long cnt = 0;

    [coder registerDefinition:(IDLDefinition*)link forLocation:-2*sizeof(long)];

    [coder decodeString:&typeId];
    [coder decodeString:&str];
    [coder decodeLong:&cnt];
    while(cnt-- > 0){
        NSString* memberName;
        IDLDefinition* memberType;
        [coder decodeString:&memberName];
        memberType = [IDLDefinition definitionFromCoder:coder];
		[array addObject:[IDLStructMember structMemberWithName:memberName ofType:memberType]];
    }
    
    obj = [self structWithName:str members:array];
    [obj setTypeId:typeId];
    existingObj = (IDLStruct*)[[IDLDefinitionScope currentScope]definitionForTypeId:typeId];
    if([existingObj isEqual:obj])
        obj = existingObj;

    if(obj){
        [link setLink:obj];
        return link;
    }else
        [NSException raise:CORBAMarshalException 
            format:@"unable to un-marshal TypeCode, at %d", [coder location]];
    
    return nil;
}

- (BOOL) isEqual:(IDLStruct*)def
{
	return def == self || [def isMemberOfClass:[self class]] && [def->_members isEqualToArray:_members];
}

@end

@implementation IDLException(CORBACoding)

+ defaultObjectClass
{
	return [NSMutableDictionary class];
}

+ (void) setDefaultObjectClass:(Class)class
{
}

@end

