/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import "IDLTaggedComponent.h"
#import "IDLEncapsulation.h"
#import "IDLDefinitionScope.h"
#import "IDLDefinition_Private.h"
#import "IDLServiceContext.h"
#import "ORB_Private.h"

@implementation IDLTaggedComponent

static int _nativeCharCode = 0x05010001;
static NSArray* _stringCharCodes = nil;
static IDLTaggedComponent* _codeSetsComponent = nil;

static NSMutableDictionary* _componentTagPaths = nil;

+ (void) initialize
{
	if([IDLTaggedComponent class] == self){
		_componentTagPaths = [[NSMutableDictionary alloc]init];
		[self registerComponentTag:TAG_CODE_SETS withIDLPath:@"CONV_FRAME/CodeSetComponentInfo"];
		_stringCharCodes = [[NSArray alloc]initWithObjects:
				[NSNumber numberWithUnsignedInt:0x00010020], // ASCII
				nil];
	}
}

+ (void) setNativeCharCode:(unsigned)code
{
	if([IDLServiceContext stringEncodingWithCode:code] != NSNotFound)
		_nativeCharCode = code;
	else
		NSLog(@"Unsupported string encoding code 0x%08x", code);
}

+ (void) setTransmissionCharCodes:(NSArray*)array
{
	if(array){
		NSEnumerator* en = [array objectEnumerator];
		id code;
		while(code = [en nextObject]){
			if(![code isKindOfClass:[NSNumber class]]){
				NSLog(@"Invalid string encoding code '%@' must be NSNumber", code);
				return;
			}else
			if([IDLServiceContext stringEncodingWithCode:[code intValue]] == NSNotFound){
				NSLog(@"Unsupported string encoding code 0x%08x", [code intValue]);
				return;
			}
		}
		_stringCharCodes = [[NSArray alloc]initWithArray:array];
	}
}

- (void) _encodeWithDefPath:(NSString*)path
{
    IDLDefinition* def = (IDLDefinition*)[[IDLDefinitionScope rootScope]definitionForNamePath:path];
	if(!def)
		def = (IDLDefinition*)[[[ORB defaultORB] rootScope]definitionForNamePath:path];
    if(def){
        IDLEncapsulation* stream = [IDLEncapsulation encapsulation];
        [def encodeObject:_component withCoder:stream];
        [_component_data release];
        _component_data = [[stream data]retain];
    }else
        [NSException raise:CORBADefinitionException format:@"Cannot find definition for IDL path '%@'", path];
}

- (void) _encodeORBType
{
    IDLEncapsulation* stream = [IDLEncapsulation encapsulation];
	unsigned long type = [_component unsignedLongValue];
	[stream encodeULong:type];
	[_component_data release];
	_component_data = [[stream data]retain];
}

- (void) _encodeAltIIOPAddress
{
	IDLEncapsulation* stream = [IDLEncapsulation encapsulation];
	[stream encodeString:[_component valueForKey:@"host"]];
	[stream encodeUShort:[[_component valueForKey:@"port"]unsignedShortValue]];
	[_component_data release];
	_component_data = [[stream data]retain];
}

- (void) encode
{
    switch(_tag){
		case TAG_ORB_TYPE:
			[self _encodeORBType];
			break;
			
		case TAG_ALTERNATE_IIOP_ADDRESS:
			[self _encodeAltIIOPAddress];
			break;
			
        default:{
			NSString* path = [_componentTagPaths objectForKey:[NSNumber numberWithUnsignedInt:_tag]];
			if(path)
				[self _encodeWithDefPath:path];
			else{
				NSLog(@"IDLTaggedComponent: no IDL path is registered for tagged component with tag %u", _tag);
				[NSException raise:CORBADefinitionException 
					format:@"no IDL path is registered for tagged component with tag %u", _tag];
			}
			
			break;
		}
    }
}

- initWithTag:(unsigned)tag component:component
{
    self = [super init];
    _tag = tag;
    _component = [component retain];
//    [self encode];
    return self;
}

+ (void) registerComponentTag:(unsigned)tag withIDLPath:(NSString*)path
{
	if(path)
		[_componentTagPaths setObject:path forKey:[NSNumber numberWithUnsignedInt:tag]];
	else
		[_componentTagPaths removeObjectForKey:[NSNumber numberWithUnsignedInt:tag]];
}

- (unsigned) componentTag
{
	return _tag;
}

+ codeSetsComponent
{
	if(!_codeSetsComponent){
		NSNumber* charNative = [NSNumber numberWithLong:_nativeCharCode];
		NSNumber* wcharNative = [NSNumber numberWithLong:0x00010109]; //UTF-16
		NSNumber* wcharConvArray = [NSArray arrayWithObject:[NSNumber numberWithLong:0x00010100]]; //UCS-2

		// char & string encoding
		NSDictionary* charValue = [NSDictionary dictionaryWithObjectsAndKeys:
			charNative, @"native_code_set",
			_stringCharCodes, @"conversion_code_sets", nil];
    
		// wchar & wstring encoding
		NSDictionary* wcharValue = [NSDictionary dictionaryWithObjectsAndKeys:
			wcharNative, @"native_code_set",
			wcharConvArray, @"conversion_code_sets", nil];
        
		NSDictionary* value = [NSDictionary dictionaryWithObjectsAndKeys:
			charValue, @"ForCharData",
			wcharValue, @"ForWcharData", nil];
		_codeSetsComponent = [[self alloc]initWithTag:TAG_CODE_SETS component:value];
	}
	return _codeSetsComponent;
}

- init
{
    return [self initWithTag:0 component:nil];
}

- (void) dealloc
{
    [_component_data release];
    [_component release];
    [super dealloc];
}

- (int) tag
{
	return _tag;
}

- component_data
{
	if(_component)
		[self encode];
	return _component_data;
}

- (void) _decodeWithDefPath:(NSString*)path
{
    IDLDefinition* def = (IDLDefinition*)[[IDLDefinitionScope rootScope]definitionForNamePath:path];
	if(!def)
		def = (IDLDefinition*)[[[ORB defaultORB] rootScope]definitionForNamePath:path];
    if(def){
        IDLEncapsulation* stream = [IDLEncapsulation encapsulationWithData:(NSMutableData*)_component_data];
        [def decodeObject:&_component withCoder:stream];
        [_component retain];
    }else{
        [NSException raise:CORBADefinitionException format:@"Cannot find definition for IDL path '%@'", path];
	}
}
- (void) _decodeORBType
{
    IDLEncapsulation* stream = [IDLEncapsulation encapsulationWithData:(NSMutableData*)_component_data];
    unsigned long type;
    [stream decodeULong:&type];
    _component = [[NSNumber numberWithUnsignedLong:type]retain];
}

- (void) _decodeAltIIOPAddress
{
    IDLEncapsulation* stream = [IDLEncapsulation encapsulationWithData:(NSMutableData*)_component_data];
    NSString* host;
    unsigned short port;
    [stream decodeString:&host];
    [stream decodeUShort:&port];
    _component = [[NSDictionary dictionaryWithObjectsAndKeys:
                host, @"host", [NSNumber numberWithUnsignedShort:port], @"port", nil]retain];
}

- (void) decode
{
    switch(_tag){
        case TAG_ORB_TYPE:
            [self _decodeORBType];
            break;
        
		case TAG_ALTERNATE_IIOP_ADDRESS:
            [self _decodeAltIIOPAddress];
        
		default:
		{
			NSString* path = [_componentTagPaths objectForKey:[NSNumber numberWithUnsignedInt:_tag]];
			if(path)
				[self _decodeWithDefPath:path];
            break;
		}
    }
}

- (void) _setComponent_data:(NSData*)data
{
    if(_component_data != data){
        [_component_data release];
        _component_data = [data retain];
        [self decode];
    }
}

- component
{
	return _component;
}

- (unsigned)charEncoding
{
	unsigned code = [[_component valueForKeyPath:@"ForCharData.native_code_set"]unsignedIntValue];
	if([IDLServiceContext stringEncodingWithCode:code] != NSNotFound)
		return code;
	else{
		NSEnumerator* en = [[_component valueForKeyPath:@"ForCharData.conversion_code_sets"]objectEnumerator];
		while(code = [[en nextObject]unsignedIntValue])
			if([IDLServiceContext stringEncodingWithCode:code] != NSNotFound)
				break;
		return code? code : 0x05010001; // UTF-8, if no match
	}
}

- (unsigned)wcharEncoding
{
	int code = [[_component valueForKeyPath:@"ForWcharData.native_code_set"]unsignedIntValue];
	if([IDLServiceContext wstringEncodingWithCode:code] != NSNotFound)
		return code;
	else{
		NSEnumerator* en = [[_component valueForKeyPath:@"ForWCharData.conversion_code_sets"]objectEnumerator];
		while(code = [[en nextObject]unsignedIntValue])
			if([IDLServiceContext wstringEncodingWithCode:code] != NSNotFound)
				break;
		return code? code : 0x00010100; // UCS-2, if no match
	}
}

@end
