/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import "ORB.h"
#import "ORB_Private.h"
#import "CORBAObject.h"
#import "CORBAObject_Private.h"
#import "VXConnection.h"
#import <IDL.h>
#import "IDLTaggedComponent.h"
#import "ORBNotificationCenter.h"

@implementation ORB

static ORB* _defaultORB = nil;
static NSZone* _zone = NULL;
static NSMutableDictionary* _services = nil;
static NSMutableDictionary* _objects = nil;
static NSMutableArray* _context = nil; // pairs 'name, value' (NSDictionary)
static IDLParser* _rootParser = nil;
static NSLock* _lock = nil;
static ORBNotificationCenter *_notificationCenter = nil;

+ (NSZone*)zone
{
    return _zone;
}

#pragma mark --- ORB notification ---

+ (void) processNotification:(NSNotification*)note
{
	@try{
		[[NSNotificationCenter defaultCenter]postNotification:note];
	}@catch(NSException *e){
		NSLog(@"Ignoring exception during notification: %@, %@", [e name], [e reason]);
	}
}

+ (void) postNotification:(NSNotification*)note
{
	[self performSelectorOnMainThread:@selector(processNotification:) withObject:note waitUntilDone:NO];
}

+ (BOOL) handlesRequestsOnMainThread
{
    return [VXConnection handlesRequestsOnMainThread];
}

+ (void) setHandlesRequestsOnMainThread:(BOOL)flag
{
	[VXConnection setHandlesRequestsOnMainThread:flag];
}

+ (unsigned) maxNumberOfInvocationThreads
{
	return [VXConnection maxNumberOfInvocationThreads];
}

+ (void) setMaxNumberOfInvocationThreads:(unsigned)num
{
	[VXConnection setMaxNumberOfInvocationThreads:num];
}

// compatibility with prev versions
+ (void) setHandlesRequestsOnMainThread
{
	[self setHandlesRequestsOnMainThread:YES];
}

#pragma mark --- Initial Services ---

+ (void) _initServices
{
	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc]init];
    NSUserDefaults* defaults = [NSUserDefaults standardUserDefaults];
    NSString* string;

    string = [defaults stringForKey:@"ORBNamingIOR"];
    if(string){
        NSData* data = [self parseIORString:string];
        id naming = [CORBAObject _objectWithIORData:data];
        [_services setObject:naming forKey:@"NameService"];
    }else{
        string = [defaults stringForKey:@"ORBNamingIORF"];
        if(string)
            string = [NSString stringWithContentsOfFile:string encoding:NSUTF8StringEncoding error:NULL];
        if(string){
            NSData* data = [self parseIORString:string];
            id naming = [CORBAObject _objectWithIORData:data];
            [_services setObject:naming forKey:@"NameService"];
        }
    }
    if(!string){
        string = [defaults stringForKey:@"ORBNamingURL"];
        if(string){
            CORBAObject* naming = [CORBAObject objectWithURL:string];
            if([string hasPrefix:@"corbaloc:"])
                [naming setInterface:(IDLInterface*)[[IDLDefinitionScope rootScope]
                    definitionForNamePath:@"CosNaming/NamingContext"]];
            [_services setObject:naming forKey:@"NameService"];
        }
    }
    
    if(!string){
        // check ORBNamingAddr
        // <addr>::= <host>[:<port>]
        string = [defaults stringForKey:@"ORBNamingAddr"];
        if(string){
            NSArray* array = [string componentsSeparatedByString:@":"];
            NSString* hostName = [array objectAtIndex:0];
            int port = 1050;
            const char* svc = "NameService";
            NSData* key = [NSData dataWithBytes:svc length:strlen(svc)];
            IDLInterface* ifObj = (IDLInterface*)[[IDLDefinitionScope rootScope]
                    definitionForNamePath:@"CosNaming/NamingContext"];
            id naming = nil;
            if([array count] > 1)
                port = [[array objectAtIndex:1]intValue];
            
            naming = [[[CORBAObject allocWithZone:[self zone]]initWithInterface:ifObj]autorelease];
            [naming addProfileWithKey:key host:hostName port:port];
            naming = [CORBAObject uniqueObjectForObject:naming];
            [_services setObject:naming forKey:@"NameService"];
            [ifObj setInterfaceConnection:naming];
        }
    }
	[pool release];
}

+ services
{
	[_lock lock];
	if(!_services){
		_services = [[NSMutableDictionary alloc]init];
		[self _initServices];
	}
	[_lock unlock];
	return _services;
}

+ (void) initialize
{
    if([ORB class] == self){
        _zone = NSCreateZone(8192,2048,YES);
		_lock = [[NSLock alloc]init];
        _rootParser = [[IDLParser rootParser]retain];
        _objects = [[NSMutableDictionary allocWithZone:[self zone]]init];
        _context = [[NSMutableArray allocWithZone:[self zone]]init];
		_notificationCenter = [[ORBNotificationCenter alloc]init];
    }
}

+ (ORBNotificationCenter*) notificationCenter;
{
	return _notificationCenter;
}

#pragma mark --- Contexts ---

+ (NSMutableArray*) defaultContext
{
    return _context;
}

+ (NSArray*) currentInvocationContext
{
    return nil;
}

#pragma mark --- ORB ---

static NSArray *_defaultIDLPaths()
{
    NSUserDefaults * defaults = [NSUserDefaults standardUserDefaults];
    NSString* idlFile = [defaults stringForKey:@"ORBidl"];
    if(idlFile)
        return [idlFile componentsSeparatedByString:@":"];
    else
        return nil;

}

static NSArray *_defaultSearchPaths()
{
    NSUserDefaults * defaults = [NSUserDefaults standardUserDefaults];
    NSString* idlPath = [defaults stringForKey:@"ORBidlpath"];
    if(idlPath)
        return [idlPath componentsSeparatedByString:@":"];
    else
        return nil;

}

+ (void) ORBInit
{

}

- (void) _initORB
{
    
}

- (void) addToSearchPath:(NSArray*)paths
{
    NSEnumerator* en = [paths objectEnumerator];
    NSString* dir;
    while(dir = [en nextObject])
        [_parser addSearchPath:dir];
}

- (void) loadFiles:(NSArray*)idlFiles
{
    [_parser loadFiles:idlFiles];
}

- initWithIDLFiles:(NSArray*)paths
{
	return [self initWithIDLFiles:paths searchPaths:nil];
}

- initWithIDLFiles:(NSArray*)paths searchPaths:(NSArray*)searchPaths
{
    self = [super init];
    _parser = [[IDLParser alloc]initByParsingIDLFiles:paths searchPaths:searchPaths];
    _rootScope = [_parser rootScope];
    _cache = [[NSMutableDictionary alloc]init];
    [self _initORB];
    return self;
}

- init
{
    return [self initWithIDLFiles:nil];
}

- (void) dealloc
{
    [_parser release];
    [_cache release];
    [super dealloc];
}

+ (ORB*) defaultORB
{
    if(!_defaultORB){
        _defaultORB = [[self allocWithZone:[self zone]]initWithIDLFiles:_defaultIDLPaths()
				searchPaths:_defaultSearchPaths()];
    }
    return _defaultORB;
}

- (IDLDefinitionScope*) rootScope
{
    return _rootScope;
}

- (void) set
{
// Set current scope
    [_rootScope set];
}

// Not autorelased local object
- (CORBAObject*) createObjectWithInterface:(IDLInterface*)ifObj
{
    [self set];
    return [[CORBAObject allocWithZone:[self zone]]initWithInterface:ifObj];
}

// Autoreleased object reference, i.e. remote object
- (CORBAObject*) objectWithIOR:(NSString*)iorString
{
    [self set];
    return [CORBAObject objectWithIOR:iorString];
}

- (CORBAObject*) objectWithURL:(NSString*)corbaurl
{
    [self set];
    return [CORBAObject objectWithURL:corbaurl];
}

- (IDLInterface*) interfaceNamed:(NSString*)path
{
    IDLInterface* obj = (IDLInterface*)[_rootScope definitionForNamePath:path];
    if([obj isKindOfClass:[IDLInterface class]])
        return obj;
    else
        [NSException raise:CORBADefinitionException
                    format:@"Invalid definition path '%@'", path];
    return nil;
}

- (IDLDefinition*) definitionNamed:(NSString*)path
{
    return (IDLDefinition*)[_rootScope definitionForNamePath:path];
}

// Format: IDL:<prefix><path>:<version>
// <path>: /module/def
// <prefix>: xorb.com
// <prefix>, if any, is incuded in the returned path
static NSString* _pathForTypeId(NSString* typeId)
{
    if([typeId hasPrefix:@"IDL:"]){
        NSRange r = {4, [typeId length] - 4};
        NSRange r1 = [typeId rangeOfString:@":" options:0 range:r];
        if(r1.length > 0)
            r.length = r1.location - r.location;
        return [typeId substringWithRange:r];
    }else
    if([typeId hasPrefix:@"RMI:"]){
        NSRange r = {4, [typeId length] - 4};
        NSRange r1 = [typeId rangeOfString:@":" options:0 range:r];
        if(r1.length > 0)
            r.length = r1.location - r.location;
        return [[[typeId substringWithRange:r]componentsSeparatedByString:@"."]
                    componentsJoinedByString:@"/"];
    }else
        return nil;
}

NSString* VXVersionFromTypeId(NSString* typeId)
{
    if([typeId hasPrefix:@"IDL:"]){
        NSRange r = [typeId rangeOfString:@":" options:NSBackwardsSearch];
        if(r.length > 0)
            return [typeId substringFromIndex:r.location + r.length];
    }else
    if([typeId hasPrefix:@"RMI:"]){
        NSRange r = {4, [typeId length] - 4};
        r = [typeId rangeOfString:@":" options:0 range:r];
        if(r.length > 0)
            return [typeId substringFromIndex:r.location + r.length];
    
    }
    return nil;
}

static NSString* _nameFromTypeId(NSString* typeId)
{
    return [_pathForTypeId(typeId) lastPathComponent];
}

+ (NSString*) nameFromTypeId:(NSString*)typeId
{
    return _nameFromTypeId(typeId);
}

- (IDLInterface*) interfaceWithTypeID:(NSString*)typeId
{
    IDLInterface* obj = (IDLInterface*)[[self rootScope]definitionForTypeId:typeId];
    if([obj isKindOfClass:[IDLInterface class]])
        return obj;
    else
        return nil;
}

- (id <IDLDefinition>) definitionWithTypeID:(NSString*)typeId
{
    return (IDLDefinition*)[[self rootScope] definitionForTypeId:typeId];
}

+ resolveInitialReferences:(NSString*)service
{
    return [[self services]objectForKey:service];
}

+ (void) run
{
    [VXConnection run];
//    [[NSRunLoop currentRunLoop]configureAsServer]
    [[NSRunLoop currentRunLoop]run];
}

+ (void) shutdown
{
    //[[NSRunLoop currentRunLoop]removePort:_orbPort forMode:NSDefaultRunLoopMode];
}


static BOOL iswhitespace(unichar x)
{
    return x == ' ' || x == '\t' || x == '\n' || x == '\r';
}
static int hextol(unichar x)
{
    if(x >='0' && x <= '9')
        return x-'0';
    else if(x >= 'a' && x <= 'f')
        return x-'a'+10;
    else if(x >= 'A' && x <= 'F')
        return x-'A'+10;
    else
        [NSException raise:@"InvalidFormatException" format:@"Not a HEX: '%c'",x];
    return -1;
        
}
+ (NSData*)parseIORString:(NSString*)iorString
{
    NSRange r = [iorString rangeOfString:@"IOR:" options:NSCaseInsensitiveSearch];
    if(r.length > 0){
        NSString* str = [iorString substringFromIndex:r.location + r.length];
        int l = [str length];
		unichar buf[l*sizeof(unichar)];
		[str getCharacters:buf];
        unichar* b = buf;

        while(l>0 && iswhitespace(b[0])){
            l -= 1; b += 1;
        }
        if(l > 0){
            int hl = 0;
            unsigned char hex[l/2+1];
            unsigned char* h = hex;
            while(l > 0 && !iswhitespace(b[0])){ // stop on whitespace
                *h++ = hextol(b[0])*16 + hextol(b[1]);
                hl += 1;
                b += 2;
                l -= 2;
            }
            return [NSData dataWithBytes:hex length:hl];
        }else
            [NSException raise:@"InvalidFormatException" format:@"Invalid IOR"];
    }
    return nil;
}

static const char* hex = "0123456789ABCDEF";

+ (NSString*) iorWithData:(NSData*)data
{
    int l = [data length];
    const unsigned char* s = [data bytes];
    char buffer[2*l+5];
    char* b = &buffer[4];
    strcpy(buffer,"IOR:");
    while(l-- > 0){
        *b++ = hex[(*s>>4)%16];
        *b++ = hex[(*s++)%16];
    }
    *b = 0;
    return [NSString stringWithUTF8String:buffer];
}

+ objectForObjectID:(NSData*)objKey
{
    return [_objects objectForKey:objKey];
}

+ (void) registerObject:(CORBAObject*)obj
{
    NSData* objKey = [obj objectKey];
    if(objKey){
        if(![_objects objectForKey:objKey]){
            [_objects setObject:obj forKey:objKey];
			// add codeset component if not added yet
			if(![obj taggedComponentWithTag:TAG_CODE_SETS])
				[obj addTaggedComponent:[IDLTaggedComponent codeSetsComponent]];
        }else
            [NSException raise:@"CORBADuplicateKey" format:@"ObjectID already in use: '%@'",
                [objKey description]];
    }
}

+ (void) unregisterObject:(CORBAObject*)obj
{
    NSData* objKey = [obj objectKey];
    if(objKey)
        [_objects removeObjectForKey:objKey];
}

+ (void) setServerTimeout:(NSTimeInterval)seconds
{
    [VXConnection setTimeout:seconds];
}

+ (NSTimeInterval) serverTimeOut
{
    return [VXConnection timeout];
}

- (NSNumber*) _enumValueNamed:(NSString*)path
{
    id cached = [_cache objectForKey:path];
    if(cached)
        return cached;
    else{
        NSString* defPath = [path stringByDeletingLastPathComponent];
        NSString* key = [path lastPathComponent];
        IDLEnum* def = (IDLEnum*)[_rootScope definitionForNamePath:defPath];
        NSNumber* value;
        if(def && [def isKindOfClass:[IDLEnum class]]){
            value = [def enumValueForKey:key];
            [_cache setObject:value forKey:path];
            return value;
        }else
            [NSException raise:IDLInvalidArgumentException 
                format:@"Cannot locate enum definition for '%@'",
                defPath];
    }
    return nil;
}
- (NSNumber*) enumValueNamed:(NSString*)path
{
    if(path)
        return [self enumValueNamed:path];
    else
        [NSException raise:IDLInvalidArgumentException format:@"NULL argument in -enumValueNamed:"];
    return nil;
}

+ (IDLDefinitionScope*) currentScope
{
    return [IDLDefinitionScope currentScope];
}

+ (void) setCurrentScope:(IDLDefinitionScope*)scope
{
    [scope set];
}

+ (BOOL) allowsIndirection
{
    return [IDLDefinition allowsIndirection];
}

// Set whether to allow using indirections when marshaling TypeCode (default NO)
+ (void) setAllowsIndirection:(BOOL)flag
{
    [IDLDefinition setAllowsIndirection:flag];
}

+ (BOOL) allowsUndefinedKeys
{
	return [IDLDefinition allowsUndefinedKeys];
}

+ (void) setAllowsUndefinedKeys:(BOOL)flag
{
	[IDLDefinition setAllowsUndefinedKeys:flag];
}

+ (void) setMapsOctetToNSData:(BOOL)flag
{
	[IDLPrimitive setMapsOctetToNSData:flag];
}
+ (BOOL) mapsOctetToNSData
{
	return [IDLPrimitive mapsOctetToNSData];
}

@end
